#!/usr/bin/python

"""
DOCUMENTATION
---
module: xml_query
version_added: "1.0"
description:
     - Retrieves facts for a file similar to the linux/unix 'chkconfig' command.
options:
  path:
    description:
      - Path to the file to operate on. File must exist ahead of time.
    required: true unless xmlstring is given
    default: null
  xmlstring:
    description:
      - A string containing XML on which to operate.
    required: true unless file is given
  xpath:
    description:
      - A valid XPath expression describing the item(s) you want to manipulate. Operates on the document root, C(/), by default.
    required: false
    default: /
    choices: []
  namespaces:
    description:
      - The namespace prefix:uri mapping for the XPath expression. Needs to be a *map*, not a list of items.
    required: false
    default: null
    choices: []


EXAMPLES
# Obtain the stats of /etc/foo.conf, and check that the file still belongs
# to 'root'. Fail otherwise.
- chkconfig:
    service: network
    runlevel:234
  register: curr_state
- fail:
    msg: "Whoops! file ownership has changed"
  when: curr_state.stat.pw_name != 'root'


RETURN
stat:
    description: dictionary containing all the stat data, some platforms might add additional fields
    returned: success
    type: dictionary
    contains:
        description:
              - service name
            required: true
            default: null
        runlevel:
            description:
              - A string containing run level numbers or none to indicate have no level
            required: true
            default: none
"""

import os
import re
import sys
import traceback

if (sys.version_info[0]) < 3 and (sys.version_info[1] < 7):
    import xml.etree.ElementTree as ET
else:
    from defusedxml import ElementTree as ET

from modules.module_base import ModuleBase


def check_value(operation, stat, item):
    if operation == 'equal':
        return stat == item
    elif operation == 'not equal':
        return not stat == item
    elif operation == 'match':
        pattern = re.compile(stat)
        match = pattern.match(item)
        return match is not None
    elif operation == 'not match':
        pattern = re.compile(stat)
        match = pattern.match(item)
        return match is None
    else:
        return False


def check_values(operation, stat, items):
    if operation == 'equals':
        for i in items:
            if i != stat:
                return False
        return True
    elif operation == 'not equal':
        for i in items:
            if i == stat:
                return False
        return True
    if operation == 'case insensitive equals':
        stat_lower = stat.lower()
        for i in items:
            if i.lower() != stat_lower:
                return False
        return True
    elif operation == 'case insensitive not equal':
        stat_lower = stat.lower()
        for i in items:
            if i.lower() == stat_lower:
                return False
        return True
    elif operation == 'match':
        pattern = re.compile(stat)
        for i in items:
            match = pattern.match(i)
            if match is None:
                return False
        return True
    elif operation == 'not match':
        pattern = re.compile(stat)
        for i in items:
            match = pattern.match(i)
            if match is not None:
                return False
        return True
    elif operation == 'include':
        stats = []
        if isinstance(stat, str):
            stats.append(stat)
        else:
            stats = stat
        for s_stat in stats:
            if s_stat in items:
                pass
            else:
                return False
        return True
    elif operation == 'exclude':
        stats = []
        if isinstance(stat, str):
            stats.append(stat)
        else:
            stats = stat
        for s_stat in stats:
            if s_stat in items:
                return False
        return True
    else:
        return False


def get_node_text(node):
    ret_str = node.text
    if ret_str == None:
        ret_str = ''
    else:
        ret_str = ret_str.strip()
    for child in node:
        ret_str = ret_str + get_node_text(child)
    return ret_str


def run(arguments):
    module = ModuleBase(
        argument_spec=dict(
            action=dict(required=False, default='audit'),
            backupfile=dict(required=False, default=None),
            path=dict(required=False, default=''),
            xmlstring=dict(required=False, default=''),
            xpath=dict(required=True),
            content=dict(required=False, default='text',
                         choices=['element', 'attribute', 'text']),
            exists=dict(required=False, default='present',
                        choices=['absent', 'present']),
            attribute=dict(required=False, default=None),
            operation=dict(required=False, default='equals',
                           choices=['equals', 'not equal',
                                    'case insensitive equals',
                                    'case insensitive not equal',
                                    'match', 'not match', 'include',
                                    'exclude']),
            value=dict(required=True),
        )
    )

    rc_flag, value = module.load_params(arguments)
    if rc_flag != 0:
        return module.fail_json(rc_flag=256, msg=value)

    # check which level need to process
    _action = module.params['action']
    _backupfile = module.params['backupfile']
    _ste_values = {}
    _ste_values['path'] = module.params['path']
    _ste_values['xmlstring'] = module.params['xmlstring']
    _ste_values['xpath'] = module.params['xpath']
    _ste_values['content'] = module.params['content']
    _ste_values['attribute'] = module.params['attribute']
    _ste_values['operation'] = module.params['operation']
    _ste_values['value'] = module.params['value']

    if _ste_values['content'] == 'attribute':
        if _ste_values['attribute'] == None or _ste_values['attribute'] == '':
            return module.fail_json(
                msg='Must set attribute when content is set as attribute.')

    _ret_actions = {'path': _ste_values['path']}

    if _action == 'harden' or _action == "rollback":
        _ret_actions = {'backupfile': _backupfile}

    if _ste_values['path'] is None and _ste_values['xmlstring'] == '':
        return module.fail_json(msg='Must set path or xmlstring.')

    xml_tree = None
    _item_values = {}
    # get current state
    try:
        if _ste_values['path'] is not None:
            if not os.path.exists(_ste_values['path']):
                return module.fail_json(
                    msg='File %s is not exists.' % _ste_values['path'])
            else:
                xml_tree = ET.parse(_ste_values['path'])
            _ret_actions['path'] = _ste_values['path']
        else:
            xml_tree = ET.fromstring(_ste_values['xmlstring'])
            _ret_actions['xmlstring'] = _ste_values['xmlstring']

    except:
        return module.fail_json(msg=traceback.format_exc())

        # find all the elements
    try:
        # check if have namespace
        _item_values['value'] = []
        root = xml_tree.getroot()
        pattern = re.compile(r'({\S+})(\S+)')
        match = pattern.match(root.tag)

        namespace = ''
        if match:
            namespace = match.groups()[0]

        xpath = _ste_values['xpath'].replace('{ns}', namespace)
        # xpath .//{ns}path could not find the root node
        tmpxpath = './/' + root.tag
        if tmpxpath == xpath:
            _item_values['value'].append(root)

        _item_values['value'].extend(xml_tree.findall(xpath))

    except:
        return module.fail_json(msg=traceback.format_exc())

    # audit and harden need to do audit
    if _action == 'audit' or _action == 'harden':
        audit_resutl = run_audit(_ret_actions, _ste_values, _item_values)

        if _action == 'audit':
            return module.exit_json(result=audit_resutl, items=_ret_actions,
                                    msg="")
        else:
            if audit_resutl == True:
                # not need to do harden
                return module.exit_json(result=True, items=_ret_actions,
                                        msg="Not need to harden.")
            else:
                return module.fail_json(msg='Support action type is audit.')

    elif _action == 'rollback':
        return module.fail_json(msg='Support action type is audit.')
    else:
        return module.fail_json(msg='Support action type is audit.')


def run_audit(ret_actions, ste_values, item_values):
    audit_resutl = True
    ret_actions['values'] = []

    if len(item_values['value']) == 0:
        value = ''
        ret_actions['values'].append(value)
        ret = check_value(ste_values['operation'], ste_values['value'], value)
        if ret == False:
            audit_resutl = False
    else:
        for node in item_values['value']:
            value = ''
            if ste_values['content'] == 'text':
                value = get_node_text(node)
            elif ste_values['content'] == 'attribute':
                if ste_values['attribute'] in node.attrib:
                    value = node.attrib[ste_values['attribute']]
                else:
                    value = ''
            else:
                value = ET.tostring(node)

            ret_actions['values'].append(value)

        ret = check_values(ste_values['operation'], ste_values['value'],
                           ret_actions['values'])
        if ret == False:
            audit_resutl = False

    return audit_resutl

